///
/// Capturing and handling of Chat events and commands.
///
/// @Author    Ole Jungclaussen, Team Fusion
/// @Version   1.0.4
///
/// ** 1.0.4 **
/// - add: dynamic help
/// 
/// You *must* add this to conf.ini and confs.ini
/// [rts]
/// ; avoids the serialization runtime error when running as server
/// ; per http://forum.1cpublishing.eu/showthread.php?t=34797
/// scriptAppDomain=0
///
/// === Example integration into Mission ===
/// //$include <put-your-path-here>/CChatCommands-vX.y.z.cs
///
/// class Mission : AMission {
///     ...
///     protected CChatCommands m_ChatCmd = null;
///     ...
///     public override void OnBattleStarted()
///     {
///         ...
///         m_ChatCmd = new CChatCommands(this);
///         m_ChatCmd.AddCommand("<tl", GetTimeLeftString, "<tl => Get time left in mission");
///         ...
///     }
///     public string GetTimeLeftString()
///     {
///         string sResponse = "Timeleft is ...";
///         return sResponse;
///     }
///
using System;
using System.Collections.Generic;  /// List
using System.Text.RegularExpressions;
using maddox.game;
using maddox.game.world;
using maddox.GP;

// needed for IPlayer:
//$reference parts/core/gamePlay.dll

public class CChatCommands {
    /// <summary>
    /// 
    /// </summary>
    public bool DEBUG_MESSAGES = false;
    /// <summary>
    /// Delegate definition (naming with class as prefix, funny errors if delegates from different classes have same name)
    /// </summary>
    public delegate void   CChatCommands_OnChatDelegate(IPlayer ifPlayerFrom, string sMsg);
    /// <summary>
    /// Delegate definition (naming with class as prefix, funny errors if delegates from different classes have same name)
    /// </summary>
    public delegate string CChatCommands_GetString();
    /// <summary>
    /// Delegate definition (naming with class as prefix, funny errors if delegates from different classes have same name)
    /// </summary>
    public delegate string CChatCommands_GetHelp(IPlayer ifPlayerFrom);

    /// Mission
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission = null;
    /// <summary>
    /// 
    /// </summary>
    protected bool m_bSerialized = false;
    /// <summary>
    /// 
    /// </summary>
    protected CChatCommands_OnChatDelegate m_dgOnChat = null;
    /// <summary>
    /// 
    /// </summary>
    protected class Command {
        /// <summary>
        /// 
        /// </summary>
        public Regex rxCmd;
        /// <summary>
        /// 
        /// </summary>
        public CChatCommands_OnChatDelegate dgOnChat = null;
        /// <summary>
        /// 
        /// </summary>
        public CChatCommands_GetString dgGetResponse = null;
        /// <summary>
        /// 
        /// </summary>
        public string sHelp;
        /// <summary>
        /// 
        /// </summary>
        public CChatCommands_GetHelp dgGetHelp = null;
    }
    /// <summary>
    /// 
    /// </summary>
    List<Command> m_aCommands = null;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CChatCommands(AMission Mission)
    {
        if(Mission.GamePlay is GameDef)
        {
            m_aCommands = new List<Command>();
            m_Mission  = Mission;
            try {
                (Mission.GamePlay as GameDef).EventChat += new GameDef.Chat(OnChat);
                m_bSerialized = true;
            }
            catch(Exception e)
            {
                if(null==e) e=null; /// silence the compiler warning
                m_Mission.GamePlay.gpLogServer(null, "***\nWARNING: CChatCommands will not work!\nAdd 'scriptAppDomain=0' to the [rts] section of conf.ini and confs.ini\n***", null);
            }
            return;
        }
        throw new System.ArgumentException("Mission.GamePlay is not GameDef", "Mission");;
    }
    /// <summary>
    /// Add a command which invokes the dgFunc
    /// </summary> 
    /// <param name="sRegExCmd">A regular expression describing the command</param>
    /// <param name="dgFunc">A "void function(IPlayer ifPlayerFrom, string sMsg)" to handle command</param>
    /// <param name="sHelp">Explanation of command</param>
    public void AddCommand(string sRegExCmd, CChatCommands_OnChatDelegate dgFunc, string sHelp)
    {
        Command Cmd = new Command();
        
        Cmd.rxCmd  = new Regex(sRegExCmd, RegexOptions.Compiled | RegexOptions.IgnoreCase);
        Cmd.dgOnChat   = dgFunc;
        Cmd.dgGetResponse = null;
        Cmd.sHelp  = sHelp;
        
        m_aCommands.Add(Cmd);
    }
    /// <summary>
    /// Add a command which invokes the dgFunc
    /// </summary> 
    /// <param name="sRegExCmd">A regular expression describing the command</param>
    /// <param name="dgFunc">A "void function(IPlayer ifPlayerFrom, string sMsg)" to handle command</param>
    /// <param name="dgGetHelp">A "string function()" that returns explanation of command</param>
    public void AddCommand(string sRegExCmd, CChatCommands_OnChatDelegate dgFunc, CChatCommands_GetHelp dgGetHelp)
    {
        Command Cmd = new Command();

        Cmd.rxCmd = new Regex(sRegExCmd, RegexOptions.Compiled | RegexOptions.IgnoreCase);
        Cmd.dgOnChat = dgFunc;
        Cmd.dgGetResponse = null;
        Cmd.dgGetHelp = dgGetHelp;

        m_aCommands.Add(Cmd);
    }
    /// <summary>
    /// Add a command that uses dgFunc to optain the response.
    /// </summary> 
    /// <param name="sRegExCmd">A regular expression describing the command</param>
    /// <param name="dgFunc">A "string function()" that returns response as string</param>
    /// <param name="sHelp">Explanation of command</param>
    public void AddCommand(string sRegExCmd, CChatCommands_GetString dgFunc, string sHelp)
    {
        Command Cmd = new Command();
        
        Cmd.rxCmd  = new Regex(sRegExCmd, RegexOptions.Compiled | RegexOptions.IgnoreCase);
        Cmd.dgOnChat   = null;
        Cmd.dgGetResponse = dgFunc;
        Cmd.sHelp  = sHelp;
        
        m_aCommands.Add(Cmd);
    }
    /// <summary>
    /// Add a command that uses dgFunc to optain the response.
    /// </summary> 
    /// <param name="sRegExCmd">A regular expression describing the command</param>
    /// <param name="dgFunc">A "string function()" that returns response as string</param>
    /// <param name="dgGetHelp">A "string function()" that returns explanation of command</param>
    public void AddCommand(string sRegExCmd, CChatCommands_GetString dgFunc, CChatCommands_GetHelp dgGetHelp)
    {
        Command Cmd = new Command();

        Cmd.rxCmd = new Regex(sRegExCmd, RegexOptions.Compiled | RegexOptions.IgnoreCase);
        Cmd.dgOnChat = null;
        Cmd.dgGetResponse = dgFunc;
        Cmd.dgGetHelp = dgGetHelp;

        m_aCommands.Add(Cmd);
    }
    /// <summary>
    /// Get List of Help on Commands.
    /// </summary> 
    public List<string> GetHelp(IPlayer ifPlayerFrom)
    {
        List<string> aHelp = new List<string>();
        foreach(Command Cmd in m_aCommands)
        {
            if(null != Cmd.dgGetHelp)
            {
                aHelp.Add(Cmd.dgGetHelp(ifPlayerFrom));
            }
            else
            {
                aHelp.Add(Cmd.sHelp);
            }
        }
        return aHelp;
    }
    /// <summary>
    /// Set this to get all chat events. Use ChatCommand.SendResponse for output.
    /// Careful: You'll get *everything*, nut just stuff that starts with '<'.
    /// </summary> 
    /// <param name="dgFunc">A function that will receive all chat events</param>
    public void SetOnChat(CChatCommands_OnChatDelegate dgFunc)
    {
        m_dgOnChat = dgFunc;
    }
    /// <summary>
    /// The chat event-handler is called *before* the message itself appears
    /// in the chat, so we must delay (default: 0.5 sec) our output - otherwise
    /// it would be shown *above* player's original message.
    /// </summary> 
    /// <param name="ifPlayerTo">The recipent of the response</param>
    /// <param name="sOutput">The response</param>
    /// <param name="dTimeout">The delay in seconds to send the response</param>
    public void SendResponse(IPlayer ifPlayerTo, string sOutput, double dTimeout=0.5)
    {
        if(0 < sOutput.Length)
        {
            m_Mission.Timeout(dTimeout, () => {
                m_Mission.GamePlay.gpLogServer(new Player[] { ifPlayerTo }, sOutput, new object[] { });
            });
        }
    }
    /// <summary>
    /// Receives, handles and/or delegates chat events.
    /// </summary> 
    /// <param name="ifPlayerFrom">The sender</param>
    /// <param name="sMsg">The chat text</param>
    protected void OnChat(IPlayer ifPlayerFrom, string sMsg)
    {
        /// if general delegate, forward as is
        if(null != m_dgOnChat) m_dgOnChat(ifPlayerFrom, sMsg);
        
        /// return here if not a command
        if(sMsg[0] != '<')
        {
            return;
        }
        
        String sLcMsg = sMsg.ToLower().Trim();
        if(DEBUG_MESSAGES)
        {
            m_Mission.GamePlay.gpLogServer(null, "CChatCommands::OnChat: received command "+sLcMsg, null);
        }

        /// Handle special commands
        switch(sLcMsg)
        {
            case "<help":
                SendResponse(ifPlayerFrom, String.Join("\n", GetHelp(ifPlayerFrom).ToArray()));
                return;
        }

        /// Handle delegated commands
        foreach(Command Cmd in m_aCommands)
        {
            if(Cmd.rxCmd.IsMatch(sLcMsg))
            {
                if(DEBUG_MESSAGES)
                {
                    m_Mission.GamePlay.gpLogServer(null, "CChatCommands::OnChat: matches against "+Cmd.rxCmd.ToString(), null);
                }
                if(null != Cmd.dgOnChat)
                {
                    Cmd.dgOnChat(ifPlayerFrom, sMsg);
                }
                else if(null != Cmd.dgGetResponse)
                {
                    SendResponse(ifPlayerFrom, Cmd.dgGetResponse());
                }
            }
        }
        
    }
}
