///
/// Manage Mission Status: Score and Time
///
/// @Author    Team Fusion
/// @Version   1.0.1
///
/// @requires
///     CArmies-v1.0.1.cs
///     CBroadcast-v1.0.1.cs
///     CTargetDamage-v1.0.1.cs
///
/// === Example integration into Mission ===
/// //$include <put-your-path-here>/CArmies-vX.y.z.cs
/// //$include <put-your-path-here>/CBroadcast-vX.y.z.cs
/// //$include <put-your-path-here>/CMissionStatus-vX.y.z.cs
/// //$include <put-your-path-here>/CTargetDamage-vX.y.z.cs
///
/// class Mission : AMission {
///     ...
///     public CMissionStatus m_MissionStatus = null;
///     ...
///     public override void OnBattleStarted()
///     {
///         ...
///         m_MissionStatus = new CMissionStatus();
///         /// Set mission time 2 hours, 30 minutes
///         m_MissionStatus.SetMissionDuration(2, 30);
///         /// Set 3 points to win mission
///         m_MissionStatus.SetPointsToWin(3);
///         ...
///         ...
///     }
///     public override void OnTickGame()
///     {
///         m_MissionStatus.OnTickGame(Time.tickCounter());
///         /// returns true if mission is already over
///         if(m_MissionStatus.IsOver())
///         {
///             return;
///         }
///         /// returns true if mission *just* now has met ending conditions
///         /// this returns true only once!
///         else if(m_MissionStatus.EndsNow())
///         {
///             Kill Mission here
///             return;
///         }
///     }
///     public string GetTimeLeftString()
///     {
///        return "Time remaining in mission: " + m_MissionStatus.GetTimeLeftString();
///     }
///     public string GetMissionScoreString()
///     {
///        return "Mission status: "+m_MissionStatus.GetScoreString(true);
///     }
///
using System;
using System.Collections.Generic;

public class CMissionStatus {
    /// Destroyed Scoreboard for one army
    public class CMissionStatus_Score {
        public int iPoints = 0;
        public List<string> aTargets = new List<string>();
    }
    /// Damage Scoreboard for one army
    public class CMissionStatus_Damage {
        public Dictionary<string, int> aDamages = new Dictionary<string, int>();
    }
    /// Mission time in ticks, there are 36 ticks in a second
    protected int m_iTimeInTicks = 0;
    /// Last tick received
    protected int m_iCurTick = 0;
    /// Number of points needed to win
    protected int m_iWinningPoints = 99;
    /// Number of points gained by armies
    protected List<CMissionStatus_Score> m_aScores = new List<CMissionStatus_Score>();
    /// List of damages gained by armies
    protected List<CMissionStatus_Damage> m_aDamages = new List<CMissionStatus_Damage>();
    /// Winner of mission
    protected int m_idxWinner = CArmies.ARMY_NONE;
    /// Status of mission (timout or winner)
    protected bool m_bEnded = false;
    /// <summary>
    /// Constructor
    /// </summary>
    public CMissionStatus()
    {
        for(int i=0; i <= CArmies.ARMY_COUNT; i++)
        {
            m_aScores.Add(new CMissionStatus_Score());
            m_aDamages.Add(new CMissionStatus_Damage());
        }
    }
    /// <summary>
    /// Set mission time as iHours plus iMinutes
    /// </summary>
    /// <param name="iHours">Number of hours</param>
    /// <param name="iMinutes">Number of minutes</param>
    public void SetMissionDuration(int iHours, int iMinutes)
    {
        m_iTimeInTicks = iHours*60*60*36 + iMinutes*60*36;
    }
    /// <summary>
    /// Set number of points needed to win.
    /// </summary>
    /// <param name="iPoints">points needed to win</param>
    public void SetPointsToWin(int iPoints)
    {
        m_iWinningPoints = iPoints;
    }
    /// <summary>
    /// Set current tick
    /// </summary>
    /// <param name="iCurTick"></param>
    public void OnTickGame(int iCurTick)
    {
        m_iCurTick = iCurTick;
    }
    /// <summary>
    /// Add achieved trget damage to score
    /// </summary>
    /// <param name="CTargetDamage"></param>
    /// <param name="idxByArmy">Army that gets the points</param>
    /// <param name="Broadcast">if supplied, Braodcast target status</param>
    public void RegisterTargetDamage(CTargetDamage TargetDamage, int idxByArmy, CBroadcast Broadcast)
    {
        if(null==TargetDamage)
        {
            return;
        }
        /// DESTROYED
        else if(TargetDamage.bDestroyed)
        {
            if(null != Broadcast)
            {
                Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, TargetDamage.sName + " has been destroyed!");
            }
            /// Give points to idxByArmy - specified army or all other armies
            for(int idxArmy=0; idxArmy <= CArmies.ARMY_COUNT; idxArmy++)
            {
                if(TargetDamage.iArmy != idxArmy)
                {
                    if((idxByArmy > 0 && idxArmy == idxByArmy) || idxByArmy <= 0)
                    {
                        m_aScores[idxArmy].iPoints += TargetDamage.iPoints;
                        m_aScores[idxArmy].aTargets.Add(TargetDamage.sName); 
                        /// remove from damage scoreboard
                        if(m_aDamages[idxArmy].aDamages.ContainsKey(TargetDamage.sName))
                        {
                            m_aDamages[idxArmy].aDamages.Remove(TargetDamage.sName);
                        }
                    }
                }
            }
        }
        /// DAMAGED
        else if(null != Broadcast)
        {
            /// update damage score board
            for(int idxArmy=0; idxArmy <= CArmies.ARMY_COUNT; idxArmy++)
            {
                if(TargetDamage.iArmy != idxArmy)
                {
                    if((idxByArmy > 0 && idxArmy == idxByArmy) || idxByArmy <= 0)
                    {
                        if(m_aDamages[idxArmy].aDamages.ContainsKey(TargetDamage.sName))
                        {
                            m_aDamages[idxArmy].aDamages[TargetDamage.sName] = TargetDamage.iPercent;
                        }
                        else
                        {
                            m_aDamages[idxArmy].aDamages.Add(TargetDamage.sName, TargetDamage.iPercent);
                        }
                    }
                }
            }
            /// Do not send this to the chat, the many messages clutter it
            Broadcast.ToScreen(CArmies.ARMY_ALL, TargetDamage.sName + " damage: "+TargetDamage.iPercent+"%");
        }
    }
    /// <summary>
    /// Check if Mission end conditions have now been met (one army has won or mission time has elapsed).
    /// This function returns true only once!
    /// </summary>
    /// <return>true if mission has just ended by one army having won or mission time having elapsed</return>
    public bool EndsNow()
    {
        /// already ended before
        if(m_bEnded)
        {
            return false;
        }
        /// check if an army won
        for(int idxArmy=1; idxArmy <= CArmies.ARMY_COUNT; idxArmy++)
        {
            if(m_aScores[idxArmy].iPoints >= m_iWinningPoints)
            {
                m_bEnded    = true;
                m_idxWinner = idxArmy;
                return true;
            }
        }
        /// check if time is over
        if(HasElapsed())
        {
            m_bEnded = true;
            return true;
        }
        return false;
    }
    /// <summary>
    /// Check if Mission is already over.
    /// </summary>
    /// <return>true if mission has ended</return>
    public bool IsOver()
    {
        return m_bEnded;
    }
    /// <summary>
    /// Get mission time left as a string
    /// </summary>
    /// <returns>"HH:MM:SS" string representation of time left in mission</return>
    public string GetTimeLeftString()
    {
        TimeSpan tsTimeLeft = GetTimeLeft();
        return string.Format("Time remaining in mission: {0:D2}:{1:D2}:{2:D2}", tsTimeLeft.Hours, tsTimeLeft.Minutes, tsTimeLeft.Seconds);
    }
    /// <summary>
    /// Get mission score status as string.
    ///  "Red has 3 points"
    ///  "Blue has 3 points"
    /// </summary>
    /// <param name="bListTargets">add list of achieved targets</param>
    /// <returns>string representation points gained by the armies</return>
    public string GetScoreString(bool bListTargets)
    {
        List<string> lstTmp = new List<string>();
        for(int idxArmy=1; idxArmy <= CArmies.ARMY_COUNT; idxArmy++)
        {
            lstTmp.Add(GetScoreString(idxArmy, bListTargets));
        }
        return String.Join("\n", lstTmp.ToArray());
    }
    /// <summary>
    /// Get army score status as string.
    /// "Red has 3 points"
    /// </summary>
    /// <param name="idxArmy">army to get the score for</param>
    /// <param name="bListTargets">add list of achieved targets</param>
    /// <returns>string representation points gained by the armies</return>
    public string GetScoreString(int idxArmy, bool bListTargets)
    {
        List<string> lstTmp = new List<string>();
        lstTmp.Add(CArmies.ARMY_COLOURS[idxArmy] + " has " + m_aScores[idxArmy].iPoints + " points");
        if(bListTargets)
        {
            /// destroyed targets
            if(m_aScores[idxArmy].iPoints > 0)
            {
                lstTmp.Add(String.Join(", ", m_aScores[idxArmy].aTargets.ToArray()));
            }
            /// damaged targets
            if(0 < m_aDamages[idxArmy].aDamages.Count)
            {
                List<string> lstTmpDmg = new List<string>();
                foreach(KeyValuePair<string, int> kvp in m_aDamages[idxArmy].aDamages)
                {
                    lstTmpDmg.Add(kvp.Key + "(" + kvp.Value.ToString() + "%)");
                }
                lstTmp.Add(String.Join(", ", lstTmpDmg.ToArray()));
            }
        }
        return String.Join("\n", lstTmp.ToArray());
    }
    /// <summary>
    /// If mission has ended, get simple result description
    /// On a draw: "The match ends in a draw! Objectives still left for both sides."
    /// On a Win:  "(Red or Blue) has successfully completed all objectives!"
    /// </summary>
    /// <returns>string containing mission result - null if mission has not ended</return>
    public string GetMissionResultString()
    {
        if(!m_bEnded)
        {
            return null;
        }
        
        return 
            CArmies.ARMY_NONE == m_idxWinner
            ? "The match ends in a draw! Objectives still left for both sides."
            : CArmies.ARMY_COLOURS[m_idxWinner] + " has successfully completed all objectives!"
        ;
    }
    /// <summary>
    /// If mission has ended, get CArmies index of winner.
    /// </summary>
    /// <returns>CArmies.ARMY_... or CArmies.ARMY_ALL if mission has not ended</return>
    public int GetWinner()
    {
        if(!m_bEnded)
        {
            return CArmies.ARMY_ALL;
        }
        return m_idxWinner;
    }
    /// <summary>
    /// Get current tick
    /// </summary>
    /// <returns>Last received tick</return>
    public int GetCurTick()
    {
        return m_iCurTick;
    }
    /// <summary>
    /// Get mission time left in ticks (one second has 36 ticks)
    /// </summary>
    /// <param name=""></param>
    public int GetTicksLeft()
    {
        return m_iTimeInTicks - m_iCurTick;
    }
    /// <summary>
    /// Get mission time left in seconds
    /// </summary>
    /// <param name=""></param>
    /// <returns>seconds left in mission</return>
    public TimeSpan GetTimeLeft()
    {
        return TimeSpan.FromSeconds(GetTicksLeft()/36);
    }
    /// <summary>
    /// Check if Mission Time has been exceeded.
    /// </summary>
    /// <returns>true if Mission Time has been reached or exceeded</return>
    public bool HasElapsed()
    {
        return m_iCurTick >= m_iTimeInTicks;
    }
}
