///
/// Handle Rearm and Refuel
///
///
/// @Author    Team Fusion
/// @Version   1.0.3
///
/// === Example integration into Mission ===
///    
///    //$include <put-your-path-here>/CChatCommands-vX.y.z.cs
///    //$include ./CMissionMenu-v1.0.0.cs
///    using MissionMenu;
///    //$include <put-your-path-here>/CRearmRefuel-vX.y.z.cs
/*
        class Mission : AMission {
            ...
            protected CRearmRefuel       m_RearmRefuel;
            ...
            public override void OnBattleStarted()
            {
                ...
                /// Initialize CRearmRefuel
                m_RearmRefuel = new CRearmRefuel(this);
                /// Add command handler(s) to CChatCommands (<rr###, where ### is between 0 and 100)
                m_RearmRefuel.AddToChatCommands(m_ChatCmd);
                /// See the CRearmRefuel.Set... functions for configuration
                ...
            }
        
*/

using System;
using System.Collections;          /// ArrayList
using System.Collections.Generic;  /// List
using System.Text.RegularExpressions;
using System.Timers;               /// Timers
using maddox.game;                 /// AMission
using maddox.game.world;           /// AiAircraft
using maddox.GP;                   /// Point3D

//$include ./CMissionMenu-v1.0.0.cs
using MissionMenu;

public class CRearmRefuel {
    /// set to true to get informative log output
    public bool DEBUG_MESSAGES = false;
    /// internal
    protected AMission     m_Mission = null;
    protected CChatCommands m_ChatCmd = null;

    protected Regex m_rxCommand = null;
    /// All requests currently pending
    protected List<RearmRefuelRequest> m_aRequests = new List<RearmRefuelRequest>();
    protected RearmRefuelRequirements m_Require = null;
    
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CRearmRefuel(AMission Mission)
    {
        m_Mission = Mission;
        m_rxCommand = new Regex(@"^<rr([0-9]+)", RegexOptions.Compiled | RegexOptions.IgnoreCase);
        m_Require = new RearmRefuelRequirements(m_Mission);
        RearmRefuelConfig.DEBUG_MESSAGES = DEBUG_MESSAGES;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="ifPlayerFrom"></param>
    /// <returns>Help text for chat command</returns>
    public string GetChatHelp(IPlayer CurPlayer)
    {
        int iMinimumFuel = !RearmRefuelConfig.bMinimumFuel ? 0 : (CurPlayer.Place() as AiAircraft).GetMinimumFuelInPercent();
        return "<rr# => Rearm and refuel; # is between "+ iMinimumFuel + " and 100";
    }
    /// <summary>
    /// Integrate with module CChatCommands
    /// </summary> 
    /// <param name="ChatCmd">The Mission's instance of CChatCommands</param>
    public void AddToChatCommands(CChatCommands ChatCmd)
    {
        m_ChatCmd = ChatCmd;
        m_ChatCmd.AddCommand(m_rxCommand.ToString(), OnChatCommand, this.GetChatHelp);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="Menu"></param>
    /// <param name="bRearmOnly"></param>
    /// <param name="aPercents"></param>
    public void AddToMissionMenu(MissionMenu.MenuItem Menu, int[] aPercents = null)
    {
        Menu.AddCommand("Rearm", OnMissionMenuCommand, new object[] { 0, true, false });
        Menu.AddCommand("Rearm guns only", OnMissionMenuCommand, new object[] { 0, true, true });

        if (null == aPercents)
        {
            RearmRefuelConfig.aMenuPercents.AddRange(new int[] { 25, 50, 75, 100 });
        }
        else
        {
            RearmRefuelConfig.aMenuPercents.AddRange(aPercents);
        }
        //foreach(int iPercent in aPercents)
        //{
        //    Menu.AddCommand("Refuel "+ iPercent.ToString()+ "%", OnMissionMenuCommand, new object[] { iPercent, false, false });
        //}
        Menu.AddDynamicMenu("Refuel", OnMissionMenuSelectRefuel_Make, OnMissionMenuSelectRefuel_Select);
    }
    /// <summary>
    /// Delegate to populate a CMissionMenu's dynamic Menu.
    /// * This is closely tied to OnMissionMenuSelectHQ_Select, as we only have the idx1Item (=the sequence) of the
    ///   menu's commands to connect to the selected command.
    ///   
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    /// </summary>
    /// <param name="Menu"></param>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuSelectRefuel_Make(MissionMenu.MenuItem Menu, Player CurPlayer, object[] aExtraData)
    {
        int iMinimumFuel = !RearmRefuelConfig.bMinimumFuel ? 0 : (CurPlayer.Place() as AiAircraft).GetMinimumFuelInPercent();
        foreach(int iPrc in RearmRefuelConfig.aMenuPercents)
        {
            if(iPrc >= iMinimumFuel)
            {
                MissionMenu.MenuItem Command = Menu.AddDynamicCommand(iPrc + "%");
                Command.m_bEnabled = true;
            }
        }
    }
    /// <summary>
    /// Delegate to process CMissionMenu's dynamic Menu commad selected.
    /// 
    /// * This is closely tied to OnMissionMenuSelectHQ_Make, as we only have the idx1Item (=the sequence) of the
    ///   menu's commands to connect to the selected command.
    ///   
    /// * ATTENTION: This might also be invoked if the command had been disabled (red color)
    /// 
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="idMenu"></param>
    /// <param name="idx1Item"></param>
    /// <returns>false if the menu command is/was actually meant to be disabled</returns>
    public bool OnMissionMenuSelectRefuel_Select(Player CurPlayer, int idMenu, int idx1Item)
    {
        int iMinimumFuel = !RearmRefuelConfig.bMinimumFuel ? 0 : (CurPlayer.Place() as AiAircraft).GetMinimumFuelInPercent();
        int idx = 0;
        foreach (int iPrc in RearmRefuelConfig.aMenuPercents)
        {
            if (iPrc >= iMinimumFuel)
            {
                idx++;
                if(idx == idx1Item)
                {
                    OnMissionMenuCommand(CurPlayer, new object[] { iPrc, false, false });
                }
            }
        }
        return false;
    }
    /// <summary>
    /// Adds a rearm request to a pending refuel request - or creates a new request
    /// </summary>
    public void AddRearmRequest(IPlayer CurPlayer, bool bGunsOnly)
    {
        Cleanup();
        LogDebugMessage("AddRearmRequest");
        foreach (RearmRefuelRequest Req in m_aRequests)
        {
            if(Req.IsPlayer(CurPlayer))
            {
                LogDebugMessage("found active request");
                Req.AddRearmRequest(bGunsOnly);
                return;
            }
        }
        RearmRefuelRequest reqNew = new RearmRefuelRequest(m_Mission.GamePlay, CurPlayer, true, bGunsOnly, false, 0);
        m_aRequests.Add(reqNew);
        return;
    }
    /// <summary>
    /// Adds a refuel request to a pending rearm request - or creates a new request
    /// </summary>
    public void AddRefuelRequest(IPlayer CurPlayer, int iPercent)
    {
        Cleanup();
        LogDebugMessage("AddRefuelRequest");
        foreach (RearmRefuelRequest Req in m_aRequests)
        {
            if(Req.IsPlayer(CurPlayer))
            {
                LogDebugMessage("found active request");
                Req.AddRefuelRequest(iPercent);
                return;
            }
        }
        RearmRefuelRequest reqNew = new RearmRefuelRequest(m_Mission.GamePlay, CurPlayer, false, false, true, iPercent);
        m_aRequests.Add(reqNew);
        return;
    }
    /// <summary>
    /// Get a list of unmet requirements.
    /// </summary>
    /// <returns>List with unmet requiremenets</returns>
    public List<string>GetUnmetRequirements(IPlayer CurPlayer)
    {
        return m_Require.GetUnmetRequirements(CurPlayer);
    }
    /// <summary>
    /// Remove all requests
    /// </summary>
    public void CancelAll()
    {
        LogDebugMessage("Cancelling all requests");
        foreach (RearmRefuelRequest Req in m_aRequests)
        {
            Req.Cancel();
        }
        m_aRequests.Clear();
    }
    /// <summary>
    /// cancel request of player
    /// </summary>
    public void CancelForPlayer(IPlayer CurPlayer)
    {
        foreach (RearmRefuelRequest Req in m_aRequests)
        {
            if(Req.IsPlayer(CurPlayer))
            {
                LogDebugMessage("Cancelling for player");
                Req.Cancel();
                break;
            }
        }
        Cleanup();
    }
    /// <summary>
    /// Get Status for a player
    /// </summary>
    public string GetStatus(IPlayer CurPlayer)
    {
        foreach(RearmRefuelRequest Req in m_aRequests)
        {
            if(Req.IsPlayer(CurPlayer))
            {
                return Req.GetStatus();
            }
        }
        return "";
    }
    /// <summary>
    /// Enable/disable the minimum fuel restriction.
    /// </summary>
    /// <param name="bToggle"></param>
    public void SetMinimumFuel(bool bToggle)
    {
        RearmRefuelConfig.bMinimumFuel = bToggle;
    }
    /// <summary>
    /// Duration in seconds for a full rearm
    /// </summary>
    /// <param name="iValue"></param>
    public void SetRearmDuration(int iSeconds)
    {
        RearmRefuelConfig.iRearmDuration = iSeconds;
    }
    /// <summary>
    /// Duration in seconds for a 100% refuel
    /// Lesser fuel amounts result in relative fractions of that.
    /// Note: see iRefuelDurationMin below
    /// </summary>
    /// <param name="iValue"></param>
    public void SetRefuelDuration(int iSeconds)
    {
        RearmRefuelConfig.iRefuelDuration = iSeconds;
    }
    /// <summary>
    /// Minimum duration in seconds for refuel.
    /// Refuel time cannot be shorter than that.
    /// Even for 10%, a truck must drive up ...
    /// </summary>
    /// <param name="iValue"></param>
    public void SetMinimumRefuelDuration(int iSeconds)
    {
        RearmRefuelConfig.iRefuelDurationMin = iSeconds;
    }
    /// <summary>
    /// Interval of "... complete in Min::Sec messages
    /// Set to 0 to disable
    /// </summary>
    /// <param name="iValue"></param>
    public void SetNotifyInterval(int iSeconds)
    {
        RearmRefuelConfig.iNotifyInterval = iSeconds;
    }
    /// <summary>
    /// Show notify on Screen, too
    /// Default is false
    /// </summary>
    /// <param name="iValue"></param>
    public void SetNotifyOnScreen(bool bVal)
    {
        RearmRefuelConfig.bNotifyOnScreen = bVal;
    }
    /// <summary>
    /// Allowable factor for within airfield perimeter.
    /// The visible airfields are sometimes bigger than the technical value configured.
    /// Even spawn points may sometimes be outside of that technical perimeter
    /// Technical airfield perimeter is multiplied with this.
    /// Default is 1.5
    /// </summary>
    /// <param name="iValue"></param>
    public void SetAirfieldParameterAdjust(int dRatio)
    {
        RearmRefuelConfig.dAirfieldParameterAdjust = dRatio;
    }
    /// <summary>
    /// Allowable deviation from airfield's elevation in meter.
    /// To check if we're actually on the ground, we need to check plane's elevation
    /// against the airfield's. Unfortunately airfield elevation is not homogenous, en
    /// extreme is Maidstone with places up to 7m different from technical elevation.
    /// Note: This is strictly not necessary, because speed=0 within airfield
    ///       perimeter would mean the same, but for user friendliness we say 
    ///       'land on airfield' first and 'set chokes' later
    /// value in meters
    /// Default is 7
    /// </summary>
    /// <param name="iValue"></param>
    public void SetAirfieldElevationAdjust(int dMeters)
    {
        RearmRefuelConfig.dAirfieldElevationAdjust = dMeters;
    }
    /// <summary>
    /// maximum engines to check (tanks, fuelcocks, magnetos, etc.)
    /// Default is 2
    /// </summary>
    /// <param name="iValue"></param>
    public void SetMaxEngineCount(int iValue)
    {
        RearmRefuelConfig.iMaxEngineCount = iValue;
    }
    /// <summary>
    /// Invoked from CChatCommands
    /// </summary> 
    /// <param name="ifPlayerFrom">The sender</param>
    /// <param name="sCmd">The chat text</param>
    protected void OnChatCommand(IPlayer ifPlayerFrom, string sCmd)
    {
        Match Rslt = m_rxCommand.Match(sCmd);
        if(!Rslt.Success)
        {
            return;
        }
        
        bool bGunsOnly = false;
        int iMinimumFuel = !RearmRefuelConfig.bMinimumFuel ? 0 : (ifPlayerFrom.Place() as AiAircraft).GetMinimumFuelInPercent();
        int iFuelPercent = 100;

        /// There's a number after <rr
        if (0 < Rslt.Groups[1].Length)
        {
            iFuelPercent = int.Parse(Rslt.Groups[1].Value);
        }
        iFuelPercent = Math.Min(100, Math.Max(iMinimumFuel, iFuelPercent));
        
        LogDebugMessage("OnChatCommand: received command <rr for " +iFuelPercent.ToString()+"% fuel");
        
        List<string>aUnmetRequirements = GetUnmetRequirements(ifPlayerFrom);
        if(0==aUnmetRequirements.Count)
        {
            AddRearmRequest(ifPlayerFrom, bGunsOnly);
            if(0 < iFuelPercent) AddRefuelRequest(ifPlayerFrom, iFuelPercent);
            m_ChatCmd.SendResponse(ifPlayerFrom, GetStatus(ifPlayerFrom));
        }
        else
        {
            m_ChatCmd.SendResponse(ifPlayerFrom, String.Join("\n", aUnmetRequirements.ToArray()));
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuCommand(Player CurPlayer, object[] aExtraData)
    {
        List<string> aUnmetRequirements = GetUnmetRequirements(CurPlayer as IPlayer);
        if (0 == aUnmetRequirements.Count)
        {
            int iPercent = (int)aExtraData[0];
            bool bRearm = (bool)aExtraData[1];
            bool bGunsOnly = (bool)aExtraData[2];

            if (bRearm)
            {
                AddRearmRequest(CurPlayer as IPlayer, bGunsOnly);
                m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, "Rearm ordered", null);
            }
            if (0 < iPercent)
            {
                AddRefuelRequest(CurPlayer as IPlayer, iPercent);
                m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, "Refuel ("+ iPercent +"%) ordered", null);
            }
        }
        else
        {
            m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, String.Join(" ", aUnmetRequirements.ToArray()), null);
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, String.Join("\n", aUnmetRequirements.ToArray()), null);
        }
    }
    /// <summary>
    /// remove all completed requests
    /// </summary>
    protected void Cleanup()
    {
        m_aRequests.RemoveAll(isRequestCompleted);
    }
    /// <summary>
    /// callback for cleanup()
    /// </summary>
    private bool isRequestCompleted(RearmRefuelRequest Req)
    {
        return(Req.IsCompleted());
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    private void LogDebugMessage(string sMsg)
    {
        if (RearmRefuelConfig.DEBUG_MESSAGES)
        {
            m_Mission.GamePlay.gpLogServer(null, "CRearmRefuel: "+sMsg, null);
        }
    }
}

/// The Rearm/Refuel Configuration
/// ---------------------------------------------------------------------------
public static class RearmRefuelConfig {
    /// Respect the minimum fuel property of the player's aircraft
    public static bool bMinimumFuel = true;
    /// <summary>
    /// 
    /// </summary>
    public static List<int> aMenuPercents = new List<int>();
    /// Duration in seconds for full rearm
    public static int iRearmDuration = 90;
    /// Duration in seconds for 100% refuel
    /// Lesser fuel amounts result in relative fractions of that.
    /// Note: see iRefuelDurationMin below
    public static int iRefuelDuration = 150;
    /// Minimum duration in seconds for refuel
    /// Refuel time cannot be shorter than that
    public static int iRefuelDurationMin = 150;
    /// Interval of "... complete in Min::Sec messages
    public static int iNotifyInterval = 20;
    /// Show notify on Screen, too
    public static bool bNotifyOnScreen = false;
    /// Allowable factor for within airfield perimeter.
    /// The visible airfields are sometimes bigger than the technical value configured.
    /// Even spawn points may sometimes be outside of that technical perimeter
    /// Technical airfield perimeter is multiplied with this.
    public static double dAirfieldParameterAdjust = 1.5;
    /// Allowable deviation from airfield's elevation in meter.
    /// To check if we're actually on the ground, we need to check plane's elevation
    /// against the airfield's. Unfortunately airfield elevation is not homogenous, en
    /// extreme is Maidstone with places up to 7m different from technical elevation.
    /// Note: This is strictly not necessary, because speed=0 within airfield
    ///       perimeter would mean the same, but for user friendliness we say 
    ///       'land on airfield' first and 'set chokes' later
    /// value in meters
    public static double dAirfieldElevationAdjust = 7;
    /// maximum engines to check (tanks, fuelcocks, magnetos, etc.)
    public static int iMaxEngineCount = 2;
    /// output debugging messages
    public static bool DEBUG_MESSAGES = false;
}

/// The Rearm/Refuel Requirements
/// ---------------------------------------------------------------------------
public class RearmRefuelRequirements {
    
    protected AMission m_Mission = null;
    
    public RearmRefuelRequirements(AMission Mission)
    {
        m_Mission = Mission;
    }
    public List<string> GetUnmetRequirements(IPlayer CurPlayer)
    {
        List<string> aUnMet = new List<string>();
        
        LogDebugMessage("Aircraft: " +(CurPlayer.Place() as AiAircraft).InternalTypeName());

        /// basic: must be landed
        if(!IsLandedOnAirfield(CurPlayer))
        {
            aUnMet.Add("Land on an airfield to rearm or refuel.");
        }
        /// basic: must be undamaged
        else if(IsDamaged(CurPlayer))
        {
            aUnMet.Add("Take another aircraft, this one is damaged.");
        }
        /// LIST: FUELCOCKS, MAGNETOS, NOT-MOVING
        else
        {
            if(!HasAllFuelCocksClosed(CurPlayer))
            {
                aUnMet.Add("close fuelcocks");
            }
            if(!HasAllMagnetosOff(CurPlayer))
            {
                aUnMet.Add("switch off magnetos");
            }
            if(IsMoving(CurPlayer))
            {
                aUnMet.Add("set chocks");
            }
        }
        return aUnMet;
    }
    /// <summary>
    /// Check if all fuelcocks are closed
    /// <returns>true if all fuel cocks are closed</returns>
    /// </summary>
    public bool HasAllFuelCocksClosed(IPlayer CurPlayer)
    {
        bool bAllClosed = true;
        AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
        double dPos = 0;

        /// The G.50 cannot close fuelcock
        if(PlayersAircraft.InternalTypeName().Contains("Aircraft.G50"))
        {
            LogDebugMessage("Checking fuelcocks, G.50 has none");
        }
        else for(int i=0; bAllClosed && i < RearmRefuelConfig.iMaxEngineCount; i++)
        {
            try
            {
                dPos = PlayersAircraft.getParameter(part.ParameterTypes.C_FuelTankSelector, i);

                LogDebugMessage("Checking fuelcocks, engine " +i+": "+dPos);

                /// Spitfire's values are reversed
                /// 0: ON, 1: OFF
                if(PlayersAircraft.InternalTypeName().Contains("Spitfire"))
                {
                    if(1!=dPos) bAllClosed=false;
                }
                /// Ju 87 B-2 values are reversed, too
                /// 0: both, 1: left, 2: right, 3: close
                else if(PlayersAircraft.InternalTypeName().Contains("Ju-87"))
                {
                    if(3!=dPos) bAllClosed=false;
                }
                /// All others have 0: OFF and 1,2,3: something not OFF
                else
                {
                    if(0!=dPos) bAllClosed=false;
                }
            }
            /// no more fuelcocks
            catch(Exception e) /// "catch(IndexOutOfRangeException e)" would be more elegant, but Develop throws a different exception
            {
                if(i==0)
                {
                    LogDebugMessage("HasAllFuelCocksClosed, exception at " +i+ ": "+e.ToString());
                }
                break;
            }
        }
        return bAllClosed;
    }
    /// <summary>
    /// Check if all magnetos are turned off
    /// </summary>
    /// <returns>true if all magnetos are turned off</returns>
    private bool HasAllMagnetosOff(IPlayer CurPlayer)
    {
        bool bAllOff = true;
        AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
        
        for(int i=0; bAllOff && i < RearmRefuelConfig.iMaxEngineCount; i++)
        {
            try
            {
                double dVal = PlayersAircraft.getParameter(part.ParameterTypes.C_Magneto, i);

                LogDebugMessage("Checking magnetos, engine " +i+": "+dVal);

                if(0!=dVal)
                {
                    bAllOff = false;
                }
            }
            /// No more magnetos
            catch (Exception e) /// "catch(IndexOutOfRangeException e)" would be more elegant, but Develop throws a different exception
            {
                if (i==0)
                {
                    LogDebugMessage("HasAllMagnetosOff, exception at " +i+ ": "+e.ToString());
                }
            }
        }
        
        return bAllOff;
    }
    /// <summary>
    /// Check for damages
    /// </summary>
    /// <returns>true if aircraft is damaged </returns>
    public bool IsDamaged(IPlayer CurPlayer)
    {
        AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
        string[] PartNames = Enum.GetNames(typeof(part.NamedDamageTypes));
        part.NamedDamageTypes[] PartVals  = (part.NamedDamageTypes[])Enum.GetValues(typeof(part.NamedDamageTypes));        
        double dDamage = 0;
        
        for(int i = 0; i < PartNames.Length; i++)
        {
            try
            {
                dDamage = PlayersAircraft.getParameter(part.ParameterTypes.M_NamedDamage, (int)(PartVals[i]));
                if(0 != dDamage)
                {
                    return true;
                }
            }
            catch (Exception e) /// "catch(IndexOutOfRangeException e)" would be more elegant, but Develop throws a different exception
            {
                LogDebugMessage("IsDamaged " +e.ToString());
            }
        }
        return false;
    }
    /// <summary>
    /// </summary>
    /// <returns>true if the plane is moving</returns>
    private bool IsMoving(IPlayer CurPlayer)
    {
        bool bIsMoving = false;
        AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
        try
        {
            double speed = 0;
            speed = PlayersAircraft.getParameter(part.ParameterTypes.Z_VelocityTAS, -1);

            LogDebugMessage("Checking speed: " +speed);
            
            if (speed > 0)
            {
                bIsMoving = true;
            }
        }
        catch (Exception e)
        {
            LogDebugMessage("IsMoving " +e.ToString());
        }
        return bIsMoving;
    }
    /// <summary>
    /// Check if aircraft is on airfield
    /// </summary>
    /// <returns>true if aircraft is on the ground on an airfield</returns>
    public bool IsLandedOnAirfield(IPlayer CurPlayer)
    {
        bool bOnAirfield = false;
        try {
            Point3d PlaneLoc = CurPlayer.Place().Pos();
            AiAirport[] aPorts = m_Mission.GamePlay.gpAirports();
            
            foreach(AiAirport Port in aPorts)
            {
                Point3d PortLoc = new Point3d(Port.Pos().x, Port.Pos().y, Port.Pos().z);

                /// Is plane within airfield perimeter? We allow factor dAirfieldParameterAdjust
                if (PlaneLoc.distance(ref PortLoc) <= Port.FieldR()*RearmRefuelConfig.dAirfieldParameterAdjust)
                {
                    /// is plane +- X meters of airfield's elevation?
                    /// see notes on RearmRefuelConfig.dAirfieldElevationAdjust
                    if(PlaneLoc.z <= PortLoc.z + RearmRefuelConfig.dAirfieldElevationAdjust && PlaneLoc.z >= PortLoc.z - RearmRefuelConfig.dAirfieldElevationAdjust)
                    {
                        bOnAirfield = true;
                        break;
                    }
                    break;
                }
            }
        }
        catch(Exception e)
        {
            bOnAirfield = true;
            LogDebugMessage("Checking on airfield, exception " +e.ToString());
        }
        
        LogDebugMessage("Checking on airfield: " +bOnAirfield);
        
        return bOnAirfield;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    private void LogDebugMessage(string sMsg)
    {
        if (RearmRefuelConfig.DEBUG_MESSAGES)
        {
            m_Mission.GamePlay.gpLogServer(null, "CRearmRefuel: " + sMsg, null);
        }
    }
}

/// A single (Rearm/Refuel) Request
/// This class handles the delay and executing of the rearm/refuel for one player
/// ---------------------------------------------------------------------------
public class RearmRefuelRequest {
    private IGamePlay m_GamePlay;
    private IPlayer   m_Player;
    
    private System.Timers.Timer m_Timer;
    
    /// message to player
    private int  m_iMessageElapsed = 0;

    /// Rearm
    private bool m_bRearmRequest  = false;
    private bool m_bRearmGunsOnly = false;
    private int  m_iRearmDuration = 0;
    private int  m_iRearmElapsed  = 0;
    private bool m_bRearmComplete = false;
    
    /// Refuel
    private bool m_bRefuelRequest  = false;
    private int  m_iRefuelPercent  = 0;
    private int  m_iRefuelDuration  = 0;
    private int  m_iRefuelElapsed  = 0;
    private bool m_bRefuelComplete = false;
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// THE DELAY FUNCTIONS
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// <summary>
    /// Constructor: Creates a request for the player with rearm, refuel, or both
    /// If this is constructed with one request (e.g. rearm) only, the other (refuel) can be added later.
    /// </summary>
    public RearmRefuelRequest(IGamePlay GamePlay, IPlayer CurPlayer, bool bRearm, bool bRearmGunsOnly, bool bRefuel, int iRefuelPercent)
    {
        m_GamePlay = GamePlay;
        m_Player   = CurPlayer;
        
        m_Timer = new System.Timers.Timer();
        m_Timer.Interval = 1000;
        m_Timer.Elapsed += onTick;
        
        m_iMessageElapsed = 0;
        
        if(bRearm)
        {
            AddRearmRequest(bRearmGunsOnly);
        }
        if(bRefuel)
        {
            AddRefuelRequest(iRefuelPercent);
        }
    }
    /// <summary>
    /// initiates rearm sequence
    /// calculates time needed, displays initial message
    /// </summary>
    public bool AddRearmRequest(bool bGunsOnly)
    {
        LogDebugMessage("AddRearmRequest");
        if (!m_bRearmRequest)
        {
            m_bRearmRequest  = true;
            m_bRearmGunsOnly = bGunsOnly;
            m_iRearmDuration  = RearmRefuelConfig.iRearmDuration;
            m_iRearmElapsed  = 0;
            m_bRearmComplete = false;

            if(!m_Timer.Enabled)
            {
                m_Timer.Start();
            }
            return true;
        }
        else
        {
            Notify("Rearm already in progress!");
        }
        return false;
    }
    /// <summary>
    /// initiates refuel sequence
    /// calculates time needed, displays initial message
    /// </summary>
    public bool AddRefuelRequest(int iPercent)
    {
        LogDebugMessage("AddRefuelRequest");
        if (!m_bRefuelRequest)
        {
            m_bRefuelRequest  = true;
            m_iRefuelPercent  = iPercent;
            m_iRefuelDuration  = Math.Max(RearmRefuelConfig.iRefuelDurationMin, (RearmRefuelConfig.iRefuelDuration * iPercent)/100);
            m_iRefuelElapsed  = 0;
            m_bRefuelComplete = false;

            /// empty the tank
            ExecRefuel(m_Player, 0);
            
            if(!m_Timer.Enabled)
            {
                m_Timer.Start();
            }        
            return true;
        }
        else
        {
            Notify("Refuel already in progress!");
        }
        return false;
    }
    /// <summary>
    /// returns true if all requests have been completed.
    /// </summary>
    public bool IsCompleted()
    {
        if(m_bRearmRequest && !m_bRearmComplete)
        {
            return false;
        }
        if(m_bRefuelRequest && !m_bRefuelComplete)
        {
            return false;
        }
        return true;
    }
    /// <summary>
    /// returns true if this requests belongs to player CurPlayer.
    /// </summary>
    public bool IsPlayer(IPlayer CurPlayer)
    {
        return(m_Player == CurPlayer);
    }
    /// <summary>
    /// Cancel this request
    /// </summary>
    public void Cancel()
    {
        LogDebugMessage("Cancelling request");
        m_Timer.Stop();
        m_bRearmRequest   = false;
        m_bRearmComplete  = false;
        m_bRefuelRequest  = false;
        m_bRefuelComplete = false;
    }
    /// <summary>
    /// Get a status of this this request
    /// </summary>
    public string GetStatus()
    {
        List<string> aMsg = new List<string>();
        if (m_bRearmRequest)
        {
            if(m_bRearmComplete)
            {
                aMsg.Add("Rearm complete");
            }
            else
            {
                TimeSpan spanTmp = new TimeSpan(0, 0, (int)(m_iRearmDuration - m_iRearmElapsed));
                aMsg.Add("Rearm complete in "+string.Format("{0:00}:{1:00}", spanTmp.Minutes, spanTmp.Seconds));
            }
        }
        if (m_bRefuelRequest)
        {
            if(m_bRefuelComplete)
            {
                aMsg.Add("Refuel ("+m_iRefuelPercent+"%) complete");
            }
            else
            {
                TimeSpan spanTmp = new TimeSpan(0, 0, (int)(m_iRefuelDuration - m_iRefuelElapsed));
                aMsg.Add("Refuel (" + m_iRefuelPercent + "%) complete in " + string.Format("{0:00}:{1:00}", spanTmp.Minutes, spanTmp.Seconds));
            }
        }
        return String.Join(", ", aMsg.ToArray());
    }
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// INTERNAL PROCESSING
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// <summary>
    /// Checks the initiated rearm/refuel sequences for status.
    /// Executes Rearm or Refuel when delay time has elapsed.
    /// Displays update messages.
    /// <summary>
    private void onTick(Object source, ElapsedEventArgs e)
    {
        ArrayList aMsg = new ArrayList();

        if (m_bRearmRequest && !m_bRearmComplete)
        {
            if(++m_iRearmElapsed >= m_iRearmDuration) {
                ExecRearm(m_Player, m_bRearmGunsOnly);
                m_bRearmComplete = true;
                aMsg.Add("Rearm complete");
            }
        }
        
        if (m_bRefuelRequest && !m_bRefuelComplete)
        {
            if(++m_iRefuelElapsed >= m_iRefuelDuration) {
                ExecRefuel(m_Player, m_iRefuelPercent);
                m_bRefuelComplete = true;
                aMsg.Add("Refuel complete");
            }
        }
        
        /// ALL DONE
        if(IsCompleted())
        {
            m_Timer.Stop();
        }
        /// NOT ALL DONE, YET
        else if(++m_iMessageElapsed >= RearmRefuelConfig.iNotifyInterval)
        {
            if(m_bRearmRequest && !m_bRearmComplete)
            {
                TimeSpan spanTmp = new TimeSpan(0, 0, (int)(m_iRearmDuration - m_iRearmElapsed));
                aMsg.Add("Rearm complete in "+string.Format("{0:00}:{1:00}", spanTmp.Minutes, spanTmp.Seconds));
            }
            if(m_bRefuelRequest && !m_bRefuelComplete)
            {
                TimeSpan spanTmp = new TimeSpan(0, 0, (int)(m_iRefuelDuration - m_iRefuelElapsed));
                aMsg.Add("Refuel complete in "+string.Format("{0:00}:{1:00}", spanTmp.Minutes, spanTmp.Seconds));
            }
            m_iMessageElapsed = 0;
        }

        if(0 != aMsg.Count)
        {
            object[] oArr = aMsg.ToArray();
            string sTxt = "";
            for (int i = 0; i < oArr.Length; i++) {
                sTxt += (string)oArr[i];
                if(i+1 < oArr.Length) sTxt += ", ";
            }
            Notify(sTxt);
        }

    }
    /// <summary>
    /// Notify Player
    /// </summary>
    private void Notify(string sTxt)
    {
        m_GamePlay.gpLogServer(new Player[] {m_Player}, sTxt, null);
        if(RearmRefuelConfig.bNotifyOnScreen)
        {
            m_GamePlay.gpHUDLogCenter(new Player[] {m_Player}, sTxt);
        }
    }
    /// <summary>
    /// executes the rearm on the player's aircraft.
    /// </summary>
    private void ExecRearm(IPlayer CurPlayer, bool bGunsOnly)
    {
        /// The player might have gone ...
        try {
            AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
            PlayersAircraft.RearmPlane(bGunsOnly);
        }
        catch(Exception e)
        {
            LogDebugMessage("Exception on ExecRearm: " +e.ToString());
        }
    }
    /// <summary>
    /// executes the refuel on the player's aircraft.
    /// </summary>
    private void ExecRefuel(IPlayer CurPlayer, int iPercent)
    {
        /// The player might have already gone ...
        try {
            AiAircraft PlayersAircraft = CurPlayer.Place() as AiAircraft;
            PlayersAircraft.RefuelPlane(iPercent);
        }
        catch(Exception e)
        {
            LogDebugMessage("Exception on ExecRefuel: "+e.ToString());
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    private void LogDebugMessage(string sMsg)
    {
        if (RearmRefuelConfig.DEBUG_MESSAGES)
        {
            m_GamePlay.gpLogServer(null, "CRearmRefuelRequest: " + sMsg, null);
        }
    }
}
