///
/// Manage Mission Targets: Airfields.
///
/// @requires
///          CTargetDamage-v1.0.1.cs
///
/// @Author    Team Fusion
/// @Version   1.0.5
/// 
/// === History ===
/// ** 1.0.5 **
/// - corrected documentation
/// 
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CTargetAirfields-vX.y.z.cs
///
/// class Mission : AMission {
///     ...
///     protected CTargetAirfields m_TargetAirfields;
///     ...
///     public override void OnBattleStarted()
///     {
///         ...
///         m_TargetAirfields = new CTargetAirfields(this);
///         /// add targets
///         m_TargetAirfields.Add(ARMY_RED, "Eastchurch", "Eastchurch", 9000, 1);
///         m_TargetAirfields.Add(ARMY_RED, "Hawkinge", "Hawkinge", 9000, 1);
///         m_TargetAirfields.Add(ARMY_BLUE, "Boulogne Alprech", "Boulogne_Alprech", 9000, 1);
///         m_TargetAirfields.Add(ARMY_BLUE, "Coquelles", "Coquelles", 9000, 1);
///         ...
///        /// Percentage of destruction assumed to have 'killed' an airfield
///         m_TargetAirfields.SetDestroyedAtPercent(100);
///         ...
///     }
///     public override void OnBombExplosion(string title, double mass, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
///     {
///         CTargetDamage Di = m_TargetAirfields.OnBombExplosion(title, mass, pos, initiator, eventArgInt);
///         
///         if(null != Di)
///         {
///             if(Di.bDestroyed)
///             {
///                 switch(Di.iArmy)
///                 {
///                     case ARMY_BLUE:
///                         m_iPointsRed += Di.iPoints;
///                         m_aDoneByRed.Add(Di.sName);
///                         break;
///                     case ARMY_RED:
///                         m_iPointsBlue += Di.iPoints;
///                         m_aDoneByBlue.Add(Di.sName);
///                         break;
///                 }
///                 m_Broadcast.ToChatAndScreen(ARMY_ALL, Di.sName + " airfield has been destroyed!");
///             }
///             else
///             {
///                 m_Broadcast.ToScreen(ARMY_ALL, Di.sName + " airfield damage: "+Di.iPercent+"%");
///             }
///         }
///     }

using System;
using System.Collections;
using System.Collections.Generic;  /// List
using maddox.game;                 /// AMission
using maddox.game.world;           /// AiActor, AiAirfield, AiBirthPlace
using maddox.GP;                   /// Point3d

public class CTargetAirfields {
    public bool DEBUG_MESSAGES = false;
    /// Percentage at which an airfiled is considered destroyed
    protected int m_iDestroyedAt = 100;
    /// the mission
    protected AMission m_Mission = null;
    /// a target
    protected class Airfield {
        public AiAirport    m_apInGame = null;
        public AiBirthPlace m_bpInGame = null;
        public string       m_sName = "";
        public int          m_iArmy = 0;
        public int          m_iKgReceived = 0;
        public int          m_iKgToDamage100 = 0;
        public int          m_iPoints = 0;
        public bool         m_bIsDestroyed = false;
    }
    /// list of targets
    protected List<Airfield> m_aAirfields = null;
    
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CTargetAirfields(AMission Mission)
    {
        m_Mission = Mission;
        m_aAirfields = new List<Airfield>();
    }
    /// <summary>
    /// Set percentage at which an airfiled is considered destroyed
    /// </summary> 
    /// <param name="iPercent">Precentage</param>
    public void SetDestroyedAtPercent(int iPercent)
    {
        m_iDestroyedAt = iPercent;
    }
    /// <summary>
    /// Add an airfield as target.
    /// </summary> 
    /// <param name="iArmy">Army the airfield belongs to</param>
    /// <param name="sAirfieldSysName">System's name of the airfield (see FMB, View, Mission Objects</param>
    /// <param name="sBirthPlaceName">Name of the Birthplace (in .mis file, [Birthplace]). Can be null when airfield is not a spawnpoint.</param>
    /// <param name="iKgToDamage100">Amount of kilogram explosives needed to damage this airfield 100%</param>
    /// <param name="iPoints">points awarded for destruction</param>
    public bool Add(int iArmy, string sAirfieldSysName, string sBirthPlaceName, int iKgToDamage100, int iPoints)
    {
        AiBirthPlace bpInGame = null;
        if(null!=sBirthPlaceName)
        {
            sBirthPlaceName = sBirthPlaceName.Trim();
            if(null==(bpInGame = GetBirthplace(sBirthPlaceName)))
            {
                m_Mission.GamePlay.gpLogServer(null, "ERROR! CTargetAirfields::Add: sBirthPlaceName '"+sBirthPlaceName+"' not found", null);
            }
        }
        
        sAirfieldSysName = sAirfieldSysName.Trim();
        AiAirport apInGame = GetAiAirport(sAirfieldSysName);
        if(null != apInGame)
        {
            Airfield Af = new Airfield();
            Af.m_apInGame       = apInGame;
            Af.m_bpInGame       = bpInGame;
            Af.m_sName          = null != bpInGame ? sBirthPlaceName : sAirfieldSysName;
            Af.m_iArmy          = iArmy;
            Af.m_iKgReceived    = 0;
            Af.m_iKgToDamage100 = iKgToDamage100;
            Af.m_bIsDestroyed   = false;
            Af.m_iPoints        = iPoints;
            
            m_aAirfields.Add(Af);
            
            if(DEBUG_MESSAGES)
            {
                m_Mission.GamePlay.gpLogServer(null, "CTargetAirfields::Add: added airfield: "+Af.m_sName, null);
            }
            
            return true;
        }
        else
        {
            m_Mission.GamePlay.gpLogServer(null, "ERROR! CTargetAirfields::Add: Airfield '"+sAirfieldSysName+"' not found", null);
        }
        return false;
    }
    /// <summary>
    /// Check and process bomb hits on targets.
    /// </summary> 
    /// <returns>A CTargetDamage object if a target is hit, null othewise</returns>
    public CTargetDamage OnBombExplosion(string title, double dMass, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
    {
        /// Walk through all our targets
        foreach(Airfield Af in m_aAirfields)
        {
            /// if not already destroyed
            if(!Af.m_bIsDestroyed)
            {
                /// 3D position of the airfield
                Point3d apPos = Af.m_apInGame.Pos();
                /// This Airfield has been hit: Bombs were within the airfield's radius
                if (pos.distance(ref apPos) < Af.m_apInGame.FieldR())
                {
                    double dOldDamage = Af.m_iKgReceived;
                    double dNewDamage = dOldDamage + dMass; 

                    double dTargetValue = Af.m_iKgToDamage100;
                    int iOldPercent  = Convert.ToInt32(Math.Ceiling((dOldDamage * 100)/dTargetValue));
                    int iNewPercent  = Convert.ToInt32(Math.Ceiling((dNewDamage * 100)/dTargetValue));

                    Af.m_iKgReceived = Convert.ToInt32(Math.Ceiling(dNewDamage));
                    
                    if (iNewPercent >= m_iDestroyedAt)
                    {
                        /// mark as destroyed
                        Af.m_bIsDestroyed = true;
                        /// disable the birthplace
                        if(null != Af.m_bpInGame)
                        {
                            Af.m_bpInGame.destroy();
                        }
                        
                        CTargetDamage Di = new CTargetDamage();
                        Di.sName      = Af.m_sName;
                        Di.iArmy      = Af.m_iArmy;
                        Di.iPercent   = iNewPercent;
                        Di.bDestroyed = true;
                        Di.iPoints    = Af.m_iPoints;
                        
                        return Di;
                    }
                    else if(iNewPercent > iOldPercent)
                    {
                        CTargetDamage Di = new CTargetDamage();
                        Di.sName      = Af.m_sName;
                        Di.iArmy      = Af.m_iArmy;
                        Di.iPercent   = iNewPercent;
                        Di.bDestroyed = false;
                        Di.iPoints    = 0;
                        return Di;
                    }
                    return null;
                }
            }
        }
        return null;
    }
    
    /// <summary>
    /// Get the game's object for an airfield
    /// </summary> 
    /// <param name="sLcName">System Name of the airfield</param>
    /// <returns>AiAirport instance or null</returns>
    protected AiAirport GetAiAirport(string sAirfieldName)
    {
        string sLcName = sAirfieldName.Trim().ToLower();
        /// walk through *all* airfields on the map and find our 'target'
        foreach (AiAirport ap in m_Mission.GamePlay.gpAirports())
        {
            if (ap.Name().ToLower().Equals(sLcName))
            {
                return ap;
            }
        }
        return null;
    }
    /// <summary>
    /// Get the game's birthplace object for an airfield
    /// </summary> 
    /// <param name="sName">name of the birthplace</param>
    /// <returns>AiBirthPlace instance or null</returns>
    protected AiBirthPlace GetBirthplace(string sBirthPlaceName)
    {
        string sLcName = sBirthPlaceName.Trim().ToLower();
        /// walk through *all* birthplaces available and find our 'target'
        foreach (AiBirthPlace bp in m_Mission.GamePlay.gpBirthPlaces())
        {
            if (bp.Name().ToLower().Equals(sLcName))
            {
                return bp;
            }
        }
        return null;
    }
}
