///
/// Manage Mission Targets: Areas.
///
/// @requires
///          CTargetDamage-v1.0.1.cs
///
/// @Author    Team Fusion - Artist
/// @Version   1.0.0
///
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CTargetAreas-vX.y.z.cs
///
/// class Mission : AMission {
///     ...
///     protected CTargetAreas m_TargetAreas;
///     ...
///     public override void OnBattleStarted()
///     {
///         ...
///         m_TargetAreas = new CTargetAreas(this);
///         /// sName
///         ///     Name of the area
///         /// dPosX
///         ///     X Position of areas center
///         /// dPosY
///         ///     Y Position of areas center
///         /// dRadius
///         ///     size of area around center position in meters
///         /// iKgToDamage100
///         ///     Set amount of kilogram explosives needed for 100% damage
///         ///     He 111 (H-2, P), Ju 88 A-1, Wellington: 2000kg
///         ///     100% of 9000 means 5 He 111s or 5 Wellington
///         /// iPoints
///         ///     Points to award for destruction of this airfield
///         m_TargetAreas.Add(CArmies.ARMY_BLUE, "Car Pool Fort Capuzzo", 264627, 130798, 300, 9000, 1);;
///         ...
///        /// Percentage of destruction assumed to have 'killed' an area
///         m_TargetAreas.SetDestroyedAtPercent(100);
///         ...
///     }
///     public override void OnBombExplosion(string title, double mass, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
///     {
///         CTargetDamage TargetDamage = null;
///         if(null != (TargetDamage = m_TargetAreas.OnBombExplosion(title, mass, pos, initiator, eventArgInt)))
///         {
///             /// using CMissionStatus
///             m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
///             /// or, if doing it all yourself:
///             //if(TargetDamage.bDestroyed)
///             //{
///             //    switch(TargetDamage.iArmy)
///             //    {
///             //        case ARMY_BLUE:
///             //            m_iPointsRed += TargetDamage.iPoints;
///             //            m_aDoneByRed.Add(TargetDamage.sName);
///             //            break;
///             //        case ARMY_RED:
///             //            m_iPointsBlue += TargetDamage.iPoints;
///             //            m_aDoneByBlue.Add(TargetDamage.sName);
///             //            break;
///             //    }
///             //    m_Broadcast.ToChatAndScreen(ARMY_ALL, TargetDamage.sName + " has been destroyed!");
///             //}
///             //else
///             //{
///             //    m_Broadcast.ToScreen(ARMY_ALL, TargetDamage.sName + " damage: "+TargetDamage.iPercent+"%");
///             //}
///         }
///         else
///         ...
///     }

using System;
using System.Collections.Generic;  /// List
using maddox.game;                 /// AMission
using maddox.game.world;           /// AiDamageInitiator
using maddox.GP;                   /// Point2d, Point3d

public class CTargetAreas {
    public bool DEBUG_MESSAGES = false;
    /// Percentage at which an airfiled is considered destroyed
    protected int m_iDestroyedAt = 100;
    /// the mission
    protected AMission m_Mission = null;
    /// a target
    protected class CArea {
        public Point2d m_ptCenter;
        /// <summary>
        /// Meter around m_ptCenter;
        /// </summary>
        public double m_dRadius;
        public string m_sName = "";
        public int m_iArmy = 0;
        public int m_iKgReceived = 0;
        public int m_iKgToDamage100 = 0;
        public int m_iPoints = 0;
        public bool m_bIsDestroyed = false;
    }
    /// list of targets
    protected List<CArea> m_aAreas = null;
    
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CTargetAreas(AMission Mission)
    {
        m_Mission = Mission;
        m_aAreas = new List<CArea>();
    }
    /// <summary>
    /// Set percentage at which an airfiled is considered destroyed
    /// </summary> 
    /// <param name="iPercent">Precentage</param>
    public void SetDestroyedAtPercent(int iPercent)
    {
        m_iDestroyedAt = iPercent;
    }
    /// <summary>
    /// Add an Area as target.
    /// </summary>
    /// <param name="iArmy">Army the airfield belongs to</param>
    /// <param name="sName">System's name of the airfield (see FMB, View, Mission Objects</param>
    /// <param name="dPosX">X Position of areas center</param>
    /// <param name="dPosY">Y Position of areas center</param>
    /// <param name="dRadius">size of area around center position in meters</param>
    /// <param name="iKgToDamage100">Amount of kilogram explosives needed to damage this airfield 100%</param>
    /// <param name="iPoints">points awarded for destruction</param>
    /// <returns></returns>
    public bool Add(int iArmy, string sName, double dPosX, double dPosY, double dRadius, int iKgToDamage100, int iPoints)
    {
        
        CArea Area  = new CArea();

        Area.m_iArmy = iArmy;
        Area.m_sName = sName.Trim();

        Area.m_ptCenter.x = dPosX;
        Area.m_ptCenter.y = dPosY;
        Area.m_dRadius = dRadius;

        Area.m_iKgReceived = 0;
        Area.m_iKgToDamage100 = iKgToDamage100;
        Area.m_bIsDestroyed = false;
        Area.m_iPoints = iPoints;
            
        m_aAreas.Add(Area);
            
        if(DEBUG_MESSAGES)
        {
            m_Mission.GamePlay.gpLogServer(null, "CTargetAreas::Add: added area: "+Area.m_sName, null);
        }
            
        return true;
    }
    /// <summary>
    /// Check and process bomb hits on targets.
    /// </summary> 
    /// <returns>A CTargetDamage object if a target is hit, null otherwise</returns>
    public CTargetDamage OnBombExplosion(string title, double dMass, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
    {
        /// 2d position of the explosion
        Point2d ptExplosion;
        ptExplosion.x = pos.x;
        ptExplosion.y = pos.y;

        /// Walk through all our targets
        foreach (CArea Area in m_aAreas)
        {
            /// if not already destroyed
            if (!Area.m_bIsDestroyed)
            {
                /// Bombs were within the areas's radius
                if (ptExplosion.distance(ref Area.m_ptCenter) <= Area.m_dRadius)
                {
                    double dOldDamage = Area.m_iKgReceived;
                    double dNewDamage = dOldDamage + dMass; 

                    double dTargetValue = Area.m_iKgToDamage100;
                    int iOldPercent  = Convert.ToInt32(Math.Ceiling((dOldDamage * 100)/dTargetValue));
                    int iNewPercent  = Convert.ToInt32(Math.Ceiling((dNewDamage * 100)/dTargetValue));

                    Area.m_iKgReceived = Convert.ToInt32(Math.Ceiling(dNewDamage));
                    
                    if (iNewPercent >= m_iDestroyedAt)
                    {
                        /// mark as destroyed
                        Area.m_bIsDestroyed = true;
                        
                        CTargetDamage Di = new CTargetDamage();
                        Di.sName      = Area.m_sName;
                        Di.iArmy      = Area.m_iArmy;
                        Di.iPercent   = iNewPercent;
                        Di.bDestroyed = true;
                        Di.iPoints    = Area.m_iPoints;
                        
                        return Di;
                    }
                    else if(iNewPercent > iOldPercent)
                    {
                        CTargetDamage Di = new CTargetDamage();
                        Di.sName      = Area.m_sName;
                        Di.iArmy      = Area.m_iArmy;
                        Di.iPercent   = iNewPercent;
                        Di.bDestroyed = false;
                        Di.iPoints    = 0;
                        return Di;
                    }
                    return null;
                }
            }
        }
        return null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    protected void LogMessage(string sMsg)
    {
        m_Mission.GamePlay.gpLogServer(null, "CTargetAreas: " + sMsg, new object[] { });
    }
}
