///
/// Manage Mission Targets: Groups
///
/// @requires
///          CArmies-v1.0.1.cs
///          CTargetDamage-v1.0.1.cs
///
/// @Author    FG28_Matkenhauser
/// @Version   1.0.1
/// @License   Public Domain
///            To the extent possible under law, FG28_Matkenhauser has waived all copyright and
///            related or neighboring rights to CTargetGroup. This work is published from Deutschland.
///
/// === Example integration into Mission ===
/// //$include <put-your-path-here>/CTargetGroups-vX.y.z.cs
/// class Mission : AMission {
///     ...
///     protected CTargetGroups m_TargetGroups;
///     ...
///     public override void OnBattleStarted()
///     {
///         ...
///         === Target Type (homogenous) Groups ===
///         m_TargetGroups = new CTargetGroups(this);
///         /// Setup groups as targets
///         /// see OnTrigger()
///         m_TargetGroups = new CTargetGroups(this);
///         /// a blue group (2 static actors; fuel storage)
///         CTargetGroup BlueFuelStorage1 = new CTargetGroup(this, CArmies.ARMY_BLUE, "Blue Fuel Storage", 1);
///         BlueFuelStorage1.AddUnitByTriggerName("BlueFuelStorage1a");
///         BlueFuelStorage1.AddUnitByTriggerName("BlueFuelStorage1b");
///         m_TargetGroups.AddGroup(BlueFuelStorage1);
///         /// a red group (2 static actors; fuel storage)
///         CTargetGroup RedFuelStorage1 = new CTargetGroup(this, CArmies.ARMY_BLUE, "Red Fuel Storage", 1);
///         RedFuelStorage1.AddUnitByTriggerName("RedFuelStorage1a");
///         RedFuelStorage1.AddUnitByTriggerName("RedFuelStorage1b");
///         m_TargetGroups.AddGroup(RedFuelStorage1);
///         ...
///     }
///     ...
///     public override void OnTrigger(int missionNumber, string shortName, bool active)
///     {
///        ...
///        /// Check if any item in a TargetGroup was triggered (TargetDamage stays null if not)
///        else if (null != (TargetDamage = m_TargetGroups.OnTrigger(missionNumber, shortName, active)))
///        {
///            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
///        }
///        ...
///     }
/// .mis file:
/// [Trigger]
///   BlueFuelStorage1a TGroundDestroyed 99 266170 140946 50
///   BlueFuelStorage1b TGroundDestroyed 99 266208 141030 50
///   RedFuelStorage1a TGroundDestroyed 99 291670 113971 50
///   RedFuelStorage1b TGroundDestroyed 99 291736 114053 50
///     
using System;
using System.Collections;
using System.Collections.Generic;
using maddox.game;
/// <summary>
/// 
/// </summary>
public class CTargetGroup {
    /// <summary>
    /// 
    /// </summary>
    public bool DEBUG_MESSAGES = false;
    /// <summary>
    /// the mission
    /// </summary>
    protected AMission m_Mission = null;
    /// <summary>
    /// Percentage at which the *Group* is considered destroyed
    /// </summary>
    protected int m_iDestroyedAt = 100;
    /// <summary>
    /// 
    /// </summary>
    protected int m_iArmy = CArmies.ARMY_NONE;
    /// <summary>
    /// 
    /// </summary>
    protected string m_sName = "Unknown Group";
    /// <summary>
    /// 
    /// </summary>
    protected int m_iPoints = 0;
    /// <summary>
    /// 
    /// </summary>
    protected List<string> m_aTriggerNames = null;
    /// <summary>
    /// 
    /// </summary>
    protected int m_nUnitsDestroyed = 0;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    /// <param name="iArmy">Army the airfield belongs to</param>
    /// <param name="sName">Name of the Group</param>
    public CTargetGroup(AMission Mission, int iArmy, string sName, int iPoints)
    {
        m_Mission = Mission;
        m_aTriggerNames = new List<string>();
        m_iArmy = iArmy;
        m_sName = sName;
        m_iPoints = iPoints;
    }
    /// <summary>
    /// Set percentage at which the Group is considered destroyed
    /// </summary> 
    /// <param name="iPercent">Precentage</param>
    public void SetDestroyedAtPercent(int iPercent)
    {
        m_iDestroyedAt = iPercent;
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public void AddUnitByTriggerName(string sTriggerName)
    {
        m_aTriggerNames.Add(sTriggerName);
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public CTargetDamage OnTrigger(int missionNumber, string shortName, bool active)
    {
        foreach(string sTriggerName in m_aTriggerNames)
        {
            if (sTriggerName.Equals(shortName) && active)
            {
                m_Mission.GamePlay.gpGetTrigger(shortName).Enable = false;
                
                CTargetDamage Di = new CTargetDamage();
                double dOldDamage = m_nUnitsDestroyed;
                double dNewDamage = ++m_nUnitsDestroyed; 

                int iOldPercent  = Convert.ToInt32(Math.Ceiling((dOldDamage * 100)/(double)(m_aTriggerNames.Count)));
                int iNewPercent  = Convert.ToInt32(Math.Ceiling((dNewDamage * 100)/(double)(m_aTriggerNames.Count)));
                
                Di.sName      = m_sName;
                Di.iArmy      = m_iArmy;
                Di.iPercent   = iNewPercent;
                Di.bDestroyed = iNewPercent >= m_iDestroyedAt;
                Di.iPoints    = Di.bDestroyed ? m_iPoints : 0;
                
                return Di;
            }
        }
        return null;
    }
}
/// <summary>
/// 
/// </summary> 
public class CTargetGroups {
    /// <summary>
    /// 
    /// </summary>
    public bool DEBUG_MESSAGES = false;
    /// <summary>
    /// the mission
    /// </summary>
    protected int m_iDestroyedAt = 100;
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission = null;
    /// <summary>
    /// 
    /// </summary>
    protected List<CTargetGroup> m_aGroups = null;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CTargetGroups(AMission Mission)
    {
        m_Mission = Mission;
        m_aGroups = new List<CTargetGroup>();
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public void AddGroup(CTargetGroup Group)
    {
        m_aGroups.Add(Group);
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public CTargetDamage OnTrigger(int missionNumber, string shortName, bool active)
    {
        foreach(CTargetGroup Group in m_aGroups)
        {
            CTargetDamage Di = Group.OnTrigger(missionNumber, shortName, active);
            if(null != Di)
            {
                return Di;
            }
        }
        return null;
    }
}