﻿///
///
///
///
///
///

using System;
using System.Collections.Generic;

using maddox.game;
using maddox.game.world;
using maddox.GP;

using System.Reflection;
using System.Text;

//$include .\CServerStats-v1.0.0.cs
using ServerStats;

public class Mission : AMission
{
    /// <summary>
    /// Helper class for collecting involved humans in a kill
    /// </summary>
    protected class HumanScore
    {
        public string m_sHuman;
        public double m_dScore;
        public HumanScore(string sHuman, double dScore)
        {
            m_sHuman = sHuman;
            m_dScore = dScore;
        }
    };
    /// <summary>
    /// Enable to get extensive log output
    /// This is also controlled by the CServerStats.ini
    /// </summary>
    protected bool m_bLogEvents = false;
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_BaseMission = null;
    /// <summary>
    /// 
    /// </summary>
    protected CServerStats m_ServerStats = null;
    /// <summary>
    /// 
    /// </summary>
    protected int m_LastUpdTimeLeft = -2000;
    /// <summary>
    /// 
    /// This is called right after Init() - and better suited (no need to pass params to base)
    /// BattleStarted is never called on sub missions.
    /// </summary>
    public override void Inited()
    {
        if (m_bLogEvents) LogMessage("Inited");
        base.Inited();
        MissionNumberListener = -1;

        m_ServerStats = new CServerStats(this);

        /// register myself with the base mission
        m_BaseMission = Battle.GetBaseMission();
        m_BaseMission.OnIntraMissionsMessage("CServerStats: register", new object[] { this });
    }
    /// <summary>
    /// Event: mission has been ended
    /// * The '..Stoped..' is not a spelling mistake here.
    /// * We end the mission with a process.kill(), so do not rely on this method
    ///   being actually called.
    /// </summary>
    public override void OnBattleStoped()
    {
        if (m_bLogEvents) LogMessage("onBattleStoped");
        base.OnBattleStoped();
        MissionNumberListener = MissionNumber;
        m_ServerStats.EndStats();
    }
    /// <summary>
    /// Handle Messages send from other missions
    /// </summary>
    /// <param name="sMsg"></param>
    /// <param name="Obj"></param>
    /// <returns></returns>
    public override object[] OnIntraMissionsMessage(string sMsg, object[] args = null)
    {
        /// NOTE: Do not log here, "UpdateMissionTimeLeft" my come 36 times per second ...
        
        /// Ordered by probability

        /// args[0] : TimeSpan, time left in mission
        if (sMsg.Equals("UpdateMissionTimeLeft"))
        {
            /// prevent updates intervals of less that one minute
            if (m_LastUpdTimeLeft + 2000 < Time.tickCounter())
            {
                m_LastUpdTimeLeft = Time.tickCounter();
                m_ServerStats.UpdateCurrentMissionTimeLeft((TimeSpan)args[0]);
            }
        }
        /// args[0] : string, path/mission.briefing, relative to CloD-UserDoc
        else if (sMsg.Equals("SetMissionFromBriefing"))
        {
            m_ServerStats.SetCurrentMissionFromBriefing(args[0] as string);
        }
        /// args[0] : string, absolute path to .ini file containing configuration
        else if (sMsg.Equals("Start"))
        {
            m_ServerStats.StartStats(args[0] as string);
            m_bLogEvents = m_ServerStats.m_bLogEvents;
        }
        return null;
    }
    ///
    /// EVENTS
    /// ---------------------------------------------------------------------------
    /// <summary>
    /// 
    /// </summary>
    /// <param name="missionNumber"></param>
    /// <param name="shortName"></param>
    /// <param name="actor"></param>
    /// <param name="damages"></param>
    public override void OnActorDead(int missionNumber, string shortName, AiActor actor, List<DamagerScore> damages)
    {
        if (m_bLogEvents) LogMessage("OnActorDead");

        /// An aircraft was destroyed
        /// If a player died within, there'll be another call with actor is AiPerson (see below)
        if (actor is AiAircraft)
        {
            if (m_bLogEvents) LogMessage("OnActorDead is Aircraft with " + damages.Count + " Damages ");

            /// collect all humans involved in damaging aircraft
            List<HumanScore> aInvolved = new List<HumanScore>();
            double dSumHumanDamage = 0;
            foreach (DamagerScore ds in damages)
            {
                if (ds.initiator.Actor != actor)
                {
                    if (ds.initiator.Player != null)
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is Aircraft ds.initiator.Player '" + ds.initiator.Player.Name() + "' score "+ ds.score);
                        string sHuman = GetHumanName(ds.initiator.Player);
                        if(null != sHuman)
                        {
                            aInvolved.Add(new HumanScore(sHuman, ds.score));
                            dSumHumanDamage += ds.score;
                        }
                    }
                    else
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is Aircraft ds.initiator.Actor " + ds.initiator.Actor.Name());
                    }
                }
            }
            /// award (fractional) kills to all humans involved
            ///       add up human infliced damages (not necessarily 100) and award relative fractional
            ///       sum human: 60
            ///       human A:   54  => (54 * 100) / 60 => 90
            ///       human B:    6  => ( 6 * 100) / 60 => 10
            ///       
            if(0 != aInvolved.Count && dSumHumanDamage > 0)
            {
                foreach (HumanScore hs in aInvolved)
                {
                    double dScore = Math.Round((hs.m_dScore * 100) / dSumHumanDamage, 0, MidpointRounding.AwayFromZero);
                    if (m_bLogEvents) LogMessage("OnActorDead is Aircraft award '"+hs.m_sHuman+"' "+dScore);
                    m_ServerStats.AddPlayerKillAir(hs.m_sHuman, dScore);
                }
            }
        }
        /// A player died
        else if (actor is AiPerson)
        {
            if (m_bLogEvents) LogMessage("OnActorDead is AiPerson");
            AiPerson aip = actor as AiPerson;
            if (aip.Player() != null)
            {
                string sHuman = GetHumanName(aip.Player());
                if(null == sHuman)
                {
                    return;
                }
                if (m_bLogEvents) LogMessage("OnActorDead PlayerDeath " + sHuman);
                m_ServerStats.AddPlayerDeath(sHuman);
            }
        }
        else if (actor is AiGroundActor)
        {
            if (m_bLogEvents) LogMessage("OnActorDead is AiGroundActor");
            AiGroundActor groundActor = actor as AiGroundActor;
            string typeString = groundActor.Type().ToString();

            /// What type of object was destroyed
            if (typeString.Contains("Ship"))
            {
                if (m_bLogEvents) LogMessage("OnActorDead is ship with " + damages.Count + " Damages ");

                /// collect all humans involved in damaging ship
                List<HumanScore> aInvolved = new List<HumanScore>();
                foreach (DamagerScore ds in damages)
                {
                    if (ds.initiator.Player != null)
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is ship ds.initiator.Player '" + ds.initiator.Player.Name() + "' score " + ds.score);
                        string sHuman = GetHumanName(ds.initiator.Player);
                        if (null != sHuman)
                        {
                            aInvolved.Add(new HumanScore(sHuman, ds.score));
                        }
                    }
                    else
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is ship ds.initiator.Actor " + ds.initiator.Actor.Name());
                    }
                }
                /// award (fractional) kills to all humans involved
                if (0 != aInvolved.Count)
                {
                    double dScore = Math.Round((double)100 / aInvolved.Count, 0, MidpointRounding.AwayFromZero);
                    foreach (HumanScore hs in aInvolved)
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is ship award '" + hs.m_sHuman + "' " + dScore);
                        m_ServerStats.AddPlayerKillSea(hs.m_sHuman, dScore);
                    }
                }
            }
            else
            {
                if (m_bLogEvents) LogMessage("OnActorDead is ground with " + damages.Count + " Damages ");
                /// Those who did it:
                foreach (DamagerScore ds in damages)
                {
                    if (ds.initiator.Player != null)
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead is ground ds.initiator.Player '" + ds.initiator.Player.Name() + "' score " + ds.score);
                        string sHuman = GetHumanName(ds.initiator.Player);
                        if (null != sHuman)
                        {
                            if (m_bLogEvents) LogMessage("OnActorDead is ground award '" + sHuman + "' 100");
                            m_ServerStats.AddPlayerKillGround(sHuman, 100);
                        }
                    }
                    else
                    {
                        if (m_bLogEvents) LogMessage("OnActorDead ds.initiator.Actor " + ds.initiator.Actor.Name());
                    }
                }
            }
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="missionNumber"></param>
    /// <param name="shortName"></param>
    /// <param name="aircraft"></param>
    public override void OnAircraftCrashLanded(int missionNumber, string shortName, AiAircraft aircraft)
    {
        string sHuman = GetHumanName(GetHuman(aircraft));
        if (null == sHuman)
        {
            return;
        }
        if (m_bLogEvents) LogMessage("OnAircraftCrashLanded "+sHuman);
        m_ServerStats.AddPlayerCrashLanded(sHuman);
    }

    public override void OnAircraftLanded(int missionNumber, string shortName, AiAircraft aircraft)
    {
        string sHuman = GetHumanName(GetHuman(aircraft));
        if (null == sHuman)
        {
            return;
        }
        if (m_bLogEvents) LogMessage("OnAircraftLanded "+sHuman);
        m_ServerStats.AddPlayerLanded(sHuman);
    }
    /// <summary>
    /// This is *not* called:
    /// - On Airspawn
    /// </summary>
    /// <param name="missionNumber"></param>
    /// <param name="shortName"></param>
    /// <param name="aircraft"></param>
    public override void OnAircraftTookOff(int missionNumber, string shortName, AiAircraft aircraft)
    {
        string sHuman = GetHumanName(GetHuman(aircraft));
        if(null==sHuman)
        {
            return;
        }
        if (m_bLogEvents) LogMessage("OnAircraftTookOff "+sHuman);
        m_ServerStats.AddPlayerTakeOff(sHuman);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="title"></param>
    /// <param name="pos"></param>
    /// <param name="initiator"></param>
    /// <param name="eventArgInt"></param>
    public override void OnBuildingKilled(string title, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
    {
        if (m_bLogEvents) LogMessage("OnBuildingKilled");
        if (initiator.Player != null)
        {
            if (m_bLogEvents) LogMessage("OnBuildingKilled initiator.Player " + initiator.Player.Name());
            string sHuman = GetHumanName(initiator.Player);
            if (null != sHuman)
            {
                if (m_bLogEvents) LogMessage("OnBuildingKilled award '" + sHuman + "' 100");
                m_ServerStats.AddPlayerKillGround(sHuman, 100.0f);
            }
        }
        else
        {
            if (m_bLogEvents) LogMessage("OnBuildingKilled initiator.Actor " + initiator.Actor.Name());
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="person"></param>
    public override void OnPersonParachuteLanded(AiPerson person)
    {
        string sHuman = GetHumanName(person.Player());
        if (null == sHuman)
        {
            return;
        }
        if (m_bLogEvents) LogMessage("OnPersonParachuteLanded "+sHuman);
        m_ServerStats.AddPlayerBailOut(sHuman);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="person"></param>
    public override void OnPersonParachuteFailed(AiPerson person)
    {
        string sHuman = GetHumanName(person.Player());
        if (m_bLogEvents) LogMessage("OnPersonParachuteFailed "+sHuman);
        m_ServerStats.AddPlayerBailOut(sHuman);
    }
    /// <summary>
    /// Called when
    /// - The Player spawns into the vehicle for the first time ("Create")
    ///   Note: OnPlaceLeave *may* be called before with same parameters (if player stayed connected
    ///   in a battle stop ... battle start sequence).
    /// - Player switches from pilot to another crew position (e.g pilot => bombardier, pilot => gunner)
    /// - Player switches from crew position to another crew position (e.g bombardier => gunner)
    /// It is *not* called
    /// - When player switches back to pilot *not* using "Release Position"
    /// - When player goes back to pilot  using "Release Position"
    /// /// </summary>
    /// <param name="player"></param>
    /// <param name="actor"></param>
    /// <param name="placeIndex"></param>
    public override void OnPlaceEnter(Player player, AiActor actor, int placeIndex)
    {
        base.OnPlaceEnter(player, actor, placeIndex);
        string sHuman = GetHumanName(player);
        if (null != sHuman)
        {
            string sVehicle = GetVehicleDescription(actor);
            CrewFunction Place = GetVehicleCrewPosition(actor, placeIndex);
            if (m_bLogEvents) LogMessage("OnPlaceEnter '" + sHuman + "', '" + player.Army()+ "', '"+ sVehicle + "', '" + Place.ToString()+ "'");
            /// Considering this buggy, we're only interested in entering as pilot/driver
            /// This also avoids network/database traffic
            if(Place==CrewFunction.Pilot)
            {
                m_ServerStats.UpdateCurrentMissionPlayerOnPlaceEnter(sHuman, GetArmyString(player.Army()), sVehicle, Place.ToString());
            }
        }
    }
    /// <summary>
    /// Called when
    /// - When player 'leaves' his vehicle (Briefing page, click on flag, say "yes")
    /// - *AFTER* a player has disconnected (if player had been still in a vehicle when exiting the server)
    /// - Player switches from crew position to another crew position (e.g bombardier => gunner); placeIndex is previous position ('bombardier')
    /// - When player goes back to pilot using "Release Position"; placeIndex is released position (e.g. gunner, bombardier)
    /// - If a player stayed connected in a battle stop ... battle start sequence, and spawns into the vehicle for
    ///   the first time ("Create"); placeIndex is 'pilot'. OnPlaceEnter is called after this with same parameters
    /// It is *not* called
    /// - Player switches from pilot to another crew position (e.g pilot => bombardier, pilot => gunner)
    /// - When player switches back to pilot *not* using "Release Position"
    /// </summary>
    /// <param name="player"></param>
    /// <param name="actor"></param>
    /// <param name="placeIndex"></param>
    public override void OnPlaceLeave(Player player, AiActor actor, int placeIndex)
    {
        base.OnPlaceLeave(player, actor, placeIndex);
        string sHuman = GetHumanName(player);
        if (null != sHuman)
        {
            string sVehicle = GetVehicleDescription(actor);
            CrewFunction Place = GetVehicleCrewPosition(actor, placeIndex);
            if (m_bLogEvents) LogMessage("OnPlaceLeave '" + sHuman + "', '" + player.Army() + "', '" + sVehicle + "', '" + Place.ToString() + "'");
            /// Considering this buggy, we're only interested in leaving pilot/driver
            /// This also avoids network/database traffic
            if (Place == CrewFunction.Pilot)
            {
                m_ServerStats.UpdateCurrentMissionPlayerOnPlaceLeave(sHuman, GetArmyString(player.Army()), sVehicle);
            }
        }
    }
    /// <summary>
    /// Called when a player connects
    /// NOTE: This is *not* invoked if player is already connected when battle starts
    /// </summary>
    /// <param name="player"></param>
    public override void OnPlayerConnected(Player player)
    {
        base.OnPlayerConnected(player);
        string sHuman = GetHumanName(player);
        if (null != sHuman)
        {
            if (m_bLogEvents) LogMessage("OnPlayerConnected " + sHuman);
            m_ServerStats.AddCurrentMissionPlayer(sHuman);
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="player"></param>
    /// <param name="diagnostic"></param>
    public override void OnPlayerDisconnected(Player player, string diagnostic)
    {
        base.OnPlayerDisconnected(player, diagnostic);
        string sHuman = GetHumanName(player);
        if (null != sHuman)
        {
            if (m_bLogEvents) LogMessage("OnPlayerDisconnected " + sHuman);
            m_ServerStats.RemoveCurrentMissionPlayer(sHuman);
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="intWhat0"></param>
    /// <param name="stationary"></param>
    /// <param name="initiator"></param>
    /// <param name="intWhat1"></param>
    public override void OnStationaryKilled(int intWhat0, GroundStationary stationary, AiDamageInitiator initiator, int intWhat1)
    {
        if (m_bLogEvents) LogMessage("OnStationaryKilled");
        if (initiator.Player != null)
        {
            if (m_bLogEvents) LogMessage("OnStationaryKilled initiator.Player " + initiator.Player.Name());
            string sHuman = GetHumanName(initiator.Player);
            if (null != sHuman)
            {
                if (m_bLogEvents) LogMessage("OnStationaryKilled award '" + sHuman + "' 100");
                m_ServerStats.AddPlayerKillGround(sHuman, 100.0f);
            }
        }
        else
        {
            if (m_bLogEvents) LogMessage("OnStationaryKilled initiator.Actor " + initiator.Actor.Name());
        }
    }
///
/// HELPER FUNCTIONS
/// ---------------------------------------------------------------------------
    /// <summary>
    /// 
    /// </summary>
    /// <param name="iArmy"></param>
    /// <returns></returns>
    protected string GetArmyString(int iArmy)
    {
        switch (iArmy)
        {
            case 1: return "red";
            case 2: return "blue";
        }
        return "none";
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="aircraft"></param>
    /// <returns></returns>
    protected Player GetHuman(AiAircraft aircraft)
    {
        for (int i = 0; i < aircraft.Places(); i++)
        {
            if (null != aircraft.Player(i))
            {
                return aircraft.Player(i);
            }
        }
        return null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="player"></param>
    /// <returns></returns>
    protected string GetHumanName(Player player)
    {
        if (player != null)
        {
            return player.Name();
        }
        return null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="actor"></param>
    /// <param name="placeIndex"></param>
    /// <returns></returns>
    protected CrewFunction GetVehicleCrewPosition(AiActor actor, int placeIndex)
    {
        if (actor is AiAircraft)
        {
            AiAircraft ac = actor as AiAircraft;
            return ac.CrewFunctionPlace(placeIndex);
        }
        else if (actor is AiGroundActor)
        {
            AiGroundActor ga = actor as AiGroundActor;
            return ga.CrewFunctionPlace(placeIndex);
        }
        return CrewFunction.Nil;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="actor"></param>
    /// <returns></returns>
    protected string GetVehicleDescription(AiActor actor)
    {
        string sName = "";
        if (actor is AiAircraft)
        {
            AiAircraft ac = actor as AiAircraft;
            sName = ac.InternalTypeName().Split(new char[] { '.' })[1];
        }
        else if (actor is AiGroundActor)
        {
            AiGroundActor groundActor = actor as AiGroundActor;
            sName = groundActor.Type().ToString();
            if (sName.Contains("Ship"))
            {
                sName = groundActor.InternalTypeName().Split(new char[] { '.' })[1];
            }
        }
        return sName;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    protected void LogMessage(string sMsg)
    {
        GamePlay.gpLogServer(null, "CServerStats(Mission): " + sMsg, new object[] { });
    }
}
